from abc import ABC, abstractmethod

from scoda.enumerations.tokenisation_flags import TokenisationFlags
from scoda.exceptions.tokenisation_exception import TokenisationException
from scoda.sequences.sequence import Sequence
from scoda.settings.settings import PPQN


class BaseTokeniser(ABC):
    TOKEN_PAD = 0
    TOKEN_START = 1
    TOKEN_STOP = 2
    TOKEN_SEPARATOR = 3
    VOCAB_SIZE = -1

    def __init__(self) -> None:
        super().__init__()

        self.flags = dict()

        self.cur_time = None
        self.cur_time_target = None
        self.cur_rest_buffer = None

        self.prv_type = None
        self.prv_note = None
        self.prv_value = None
        self.prv_numerator = None

        """The maximum value up to which consecutive rests will be consolidated"""
        self.set_max_rest_value = PPQN

        self.reset()

    def reset(self) -> None:
        self.reset_time()
        self.reset_previous()

    def reset_time(self) -> None:
        self.cur_time = 0
        self.cur_time_target = 0
        self.cur_rest_buffer = 0

    def reset_previous(self) -> None:
        self.prv_type = None
        self.prv_note = None
        self.prv_value = -1
        self.prv_numerator = -1

    def _general_tokenise_flush_time_buffer(self, time: int, index_time_def: int) -> list[int]:
        tokens = []

        while time > self.set_max_rest_value:
            tokens.append(int(self.set_max_rest_value + index_time_def - 1))
            time -= self.set_max_rest_value

        if time > 0:
            tokens.append(int(time + index_time_def - 1))

        return tokens

    def _notelike_tokenise_flush_rest_buffer(self, apply_target: bool, wait_token: int, index_time_def: int) -> list[
        int]:
        tokens = []

        # Insert rests of length up to `set_max_rest_value`
        while self.cur_rest_buffer > self.set_max_rest_value:
            if not (self.prv_value == self.set_max_rest_value and self.flags.get(TokenisationFlags.RUNNING_VALUE,
                                                                                 False)):
                tokens.append(int(self.set_max_rest_value + index_time_def - 1))
                self.prv_value = self.set_max_rest_value

            tokens.append(int(wait_token))
            self.cur_time += self.set_max_rest_value
            self.cur_rest_buffer -= self.set_max_rest_value

        # Insert rests smaller than `set_max_rest_value`
        if self.cur_rest_buffer > 0:
            if not (self.prv_value == self.cur_rest_buffer and self.flags.get(TokenisationFlags.RUNNING_VALUE, False)):
                tokens.append(int(self.cur_rest_buffer + index_time_def - 1))
                self.prv_value = self.cur_rest_buffer
            tokens.append(int(wait_token))

        self.cur_time += self.cur_rest_buffer
        self.cur_rest_buffer = 0

        # If there are open notes, extend the sequence to the minimum needed time target
        if apply_target and self.cur_time_target > self.cur_time:
            self.cur_rest_buffer += self.cur_time_target - self.cur_time
            tokens.extend(
                self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=wait_token,
                                                          index_time_def=index_time_def))

        return tokens

    def _gridlike_tokenise_flush_grid_buffer(self, min_grid_size: int, wait_token: int) -> list[int]:
        tokens = []

        while self.cur_rest_buffer > 0:
            tokens.append(wait_token)
            self.cur_rest_buffer -= min_grid_size

        return tokens

    @abstractmethod
    def tokenise(self, sequence: Sequence, insert_trailing_separator_token: bool = True,
                 insert_border_tokens: bool = False) -> list[int]:
        pass

    @staticmethod
    @abstractmethod
    def detokenise(tokens: list[int]) -> Sequence:
        pass

    @staticmethod
    def _time_signature_to_eights(numerator: int, denominator: int) -> int:
        scaled = numerator * (8 / denominator)

        if (not float(scaled).is_integer()):
            raise TokenisationException(
                f"Original time signature of {int(numerator)}/{int(denominator)} cannot be represented as multiples of eights")

        if not 2 <= scaled <= 16:
            raise TokenisationException(
                f"Invalid time signature numerator: {scaled}")

        return int(scaled)



from __future__ import annotations

import copy

from scoda.elements.bar import Bar
from scoda.enumerations.message_type import MessageType
from scoda.exceptions.track_exception import TrackException
from scoda.sequences.sequence import Sequence


class Track:
    """Class representing a track, which is made of (multiple) bars.
    """

    def __init__(self, bars: [Bar], name: str = None) -> None:
        super().__init__()
        self.name = name
        self.bars = bars
        self.program = None

        seq = Bar.to_sequence(bars)
        program_changes = [msg for msg in seq.rel.messages if msg.message_type == MessageType.PROGRAM_CHANGE]
        if len(program_changes) > 0:
            if not all(msg.program == program_changes[0].program for msg in program_changes):
                raise TrackException("Type of instrument inconsistent")
            self.program = program_changes[0].program

    def __copy__(self) -> Track:
        bars = [copy.copy(bar) for bar in self.bars]

        cpy = Track(bars, self.name)
        return cpy

    def to_sequence(self) -> Sequence:
        return Bar.to_sequence(self.bars)



from __future__ import annotations

import copy

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.exceptions.bar_exception import BarException
from scoda.misc.music_theory import Key
from scoda.sequences.sequence import Sequence
from scoda.settings.settings import PPQN


class Bar:
    """Class representing a single bar, its length defined by a time signature."""

    def __init__(self, sequence: Sequence, numerator: int, denominator: int, key=None) -> None:
        super().__init__()

        self.sequence: Sequence = sequence
        self.time_signature_numerator = numerator
        self.time_signature_denominator = denominator
        self.key_signature = key

        # Adjust sequence
        self.sequence.normalise()

        # Assert bar has correct capacity
        if self.sequence.get_sequence_duration_relation() > self.time_signature_numerator * PPQN / (
                self.time_signature_denominator / 4):
            raise BarException("Bar capacity exceeded")

        # Pad bar
        if self.sequence.get_sequence_duration_relation() < self.time_signature_numerator * PPQN / (
                self.time_signature_denominator / 4):
            self.sequence.pad(self.time_signature_numerator * PPQN / (self.time_signature_denominator / 4))

        # Assert time signature is consistent
        relative_sequence = self.sequence.rel
        time_signatures = [msg for msg in relative_sequence.messages if
                           msg.message_type == MessageType.TIME_SIGNATURE]

        if len(time_signatures) > 1:
            raise BarException("Too many time signatures in a bar")
        if not all(msg.numerator == self.time_signature_numerator and msg.denominator == self.time_signature_denominator
                   for msg in time_signatures):
            raise BarException("Time signatures not uniform")

        # Set time signature and remove all other time signature messages
        relative_sequence = self.sequence.rel
        relative_sequence.messages = [msg for msg in relative_sequence.messages if
                                      msg.message_type != MessageType.TIME_SIGNATURE]
        relative_sequence.messages.insert(0, Message(message_type=MessageType.TIME_SIGNATURE,
                                                     numerator=self.time_signature_numerator,
                                                     denominator=self.time_signature_denominator))

        self.sequence._abs_stale = True

    def __copy__(self) -> Bar:
        bar = Bar(copy.copy(self.sequence), self.time_signature_numerator, self.time_signature_denominator,
                  self.key_signature)

        return bar

    def difficulty(self, key_signature: Key = None) -> float:
        if key_signature is None:
            key_signature = self.key_signature
        return self.sequence.difficulty(key_signature)

    def is_empty(self) -> bool:
        return self.sequence.is_empty()

    def transpose(self, transpose_by: int) -> bool:
        """See `scoda.sequence.relative_sequence.RelativeSequence.transpose`.
        """
        if self.key_signature is not None:
            self.key_signature = Key.transpose_key(self.key_signature, transpose_by)

        return self.sequence.transpose(transpose_by)

    @staticmethod
    def to_sequence(bars: [Bar]) -> Sequence:
        sequence = Sequence()
        sequences = []

        for bar in bars:
            sequences.append(bar.sequence)
        sequence.concatenate(sequences)

        return sequence



from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import mido

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.midi.midi_track import MidiTrack
from scoda.misc.scoda_logging import get_logger
from scoda.settings.settings import PPQN

if TYPE_CHECKING:
    from scoda.sequences.sequence import Sequence


class MidiFile:
    LOGGER = get_logger(__name__)

    def __init__(self) -> None:
        super().__init__()
        self.tracks: [MidiTrack] = []
        self.PPQN = PPQN

    @staticmethod
    def open(filename) -> MidiFile:
        midi_file = MidiFile()
        mido_midi_file = mido.MidiFile(filename)
        midi_file.parse_mido(mido_midi_file)
        return midi_file

    def parse_mido(self, mido_midi_file):
        self.PPQN = mido_midi_file.ticks_per_beat
        for mido_track in mido_midi_file.tracks:
            self.tracks.append(MidiTrack.parse_mido_track(mido_track))

    def convert(self, track_indices: [[int]], meta_track_indices: [int], meta_track_index: int = 0) -> list[Sequence]:
        """ Parses this `MidiFile` and returns a list of `scoda.Sequence`.

        Args:
            track_indices: A list of grouped indices of tracks. Tracks in the same group will be merged to a single
                sequence.
            meta_track_indices: A list of indices of tracks, which can contain meta messages to consider.
            meta_track_index: The index of the final sequence that should contain meta messages.

        Returns: A list of parsed sequence.

        """
        from scoda.sequences.sequence import Sequence

        # Create sequence to fill
        sequences = [[Sequence() for _ in indices] for indices in track_indices]
        meta_sequence = Sequence()

        # PPQN scaling
        scaling_factor = PPQN / self.PPQN

        # Iterate over all tracks contained in this file
        for i, track in enumerate(self.tracks):
            # Skip tracks not specified
            if not any(i in indices for indices in track_indices) and i not in meta_track_indices:
                continue

            # Keep track of current point in time
            current_point_in_time = 0

            # Get current sequence
            current_sequence = None
            if any(i in indices for indices in track_indices):
                group_indices = next(array for array in track_indices if i in array)
                current_sequence = sequences[track_indices.index(group_indices)][group_indices.index(i)]
            elif i in meta_track_indices:
                current_sequence = meta_sequence

            # Parse messages
            for j, msg in enumerate(track.messages):
                # Add time induced by message
                current_point_in_time += (msg.time * scaling_factor)
                rounded_point_in_time = round(current_point_in_time)

                # Note On
                if msg.message_type == MessageType.NOTE_ON and any(i in indices for indices in track_indices):
                    current_sequence.add_absolute_message(
                        Message(message_type=MessageType.NOTE_ON, note=msg.note, velocity=msg.velocity,
                                time=rounded_point_in_time))
                # Note Off
                elif msg.message_type == MessageType.NOTE_OFF and any(i in indices for indices in track_indices):
                    current_sequence.add_absolute_message(
                        Message(message_type=MessageType.NOTE_OFF, note=msg.note, time=rounded_point_in_time))
                # Time Signature
                elif msg.message_type == MessageType.TIME_SIGNATURE:
                    if i not in meta_track_indices:
                        MidiFile.LOGGER.debug("MidiFile: Encountered time signature change in unexpected track.")

                    meta_sequence.add_absolute_message(
                        Message(message_type=MessageType.TIME_SIGNATURE, numerator=msg.numerator,
                                denominator=msg.denominator, time=rounded_point_in_time))
                # Key Signature
                elif msg.message_type == MessageType.KEY_SIGNATURE:
                    if i not in meta_track_indices:
                        MidiFile.LOGGER.debug("MidiFile: Encountered key signature change in unexpected track.")

                    meta_sequence.add_absolute_message(
                        Message(message_type=MessageType.KEY_SIGNATURE, key=msg.key, time=rounded_point_in_time))
                # Control change
                elif msg.message_type == MessageType.CONTROL_CHANGE:
                    meta_sequence.add_absolute_message(
                        Message(message_type=MessageType.CONTROL_CHANGE, velocity=msg.velocity, control=msg.control,
                                time=rounded_point_in_time))
                # Program change
                elif msg.message_type == MessageType.PROGRAM_CHANGE:
                    current_sequence.add_absolute_message(
                        Message(message_type=MessageType.PROGRAM_CHANGE, program=msg.program,
                                time=rounded_point_in_time))
                # Unknown, e.g. MetaMessage (will be ignored)
                else:
                    pass

        merged_sequences = []

        # Merge sequence according to groups
        for sequences_to_merge in sequences:
            for seq in sequences_to_merge:
                seq.normalise()
            track = copy.copy(sequences_to_merge[0])
            track.merge(sequences_to_merge[1:])
            merged_sequences.append(track)

        if 0 > meta_track_index or meta_track_index >= len(merged_sequences):
            raise ValueError("Invalid meta track index")

        meta_track = merged_sequences[meta_track_index]
        meta_track.merge([meta_sequence])

        # Set standard time if not set
        if not any(timing_tuple[0] == 0 for timing_tuple in
                   meta_track.get_message_timings_of_type([MessageType.TIME_SIGNATURE])):
            meta_track.add_absolute_message(
                Message(message_type=MessageType.TIME_SIGNATURE, numerator=4, denominator=4, time=0))

        return merged_sequences

    def save(self, path):
        mido_midi_file = mido.MidiFile()
        mido_midi_file.ticks_per_beat = PPQN

        for track in self.tracks:
            mido_midi_file.tracks.append(track.to_mido_track())

        mido_midi_file.save(path)



import math

import numpy as np

from scoda.elements.message import Message
from scoda.settings.settings import VELOCITY_MAX, VELOCITY_BINS, PPQN


def bin_from_velocity(velocity: int) -> int:
    """Returns the bin defined by the velocity.

    Args:
        velocity: The velocity to sort into a bin

    Returns: The corresponding bin

    """
    if velocity <= 0 or velocity > VELOCITY_MAX:
        raise ValueError("Velocity not contained in any bag")

    bin_size = round(VELOCITY_MAX / VELOCITY_BINS)
    bins = [min(VELOCITY_MAX, ((i + 1) * bin_size) + bin_size / 2) for i in range(0, VELOCITY_BINS)]

    return np.digitize(velocity, bins, right=True).item(-1)


def binary_insort(collection: list, message: Message) -> None:
    """Sorts the given message into the correct position in the already sorted list.

    Args:
        collection: A list of messages that is already sorted
        message: The message to insert

    """
    lo = 0
    hi = len(collection)

    while lo < hi:
        mid = (lo + hi) // 2
        if message.time < collection[mid].time:
            hi = mid
        else:
            lo = mid + 1

    collection.insert(lo, message)


def digitise_velocity(velocity_unquantised: int) -> int:
    """Digitises velocity to bins.

    Digitises the given velocity based on the settings of scoda. Returns a value that corresponds to one of the bins,
    but not the index of the bin itself. E.g., the value of 33 could be quantised to 32 with bins of size 16.
    In this case, the value of 32 would be returned, rather than the index 2 (of the second bin).

    Args:
        velocity_unquantised: The initial velocity

    Returns: The quantised value

    """
    if velocity_unquantised == 0:
        return velocity_unquantised

    return velocity_from_bin(bin_from_velocity(velocity_unquantised))


def find_minimal_distance(element, collection) -> int:
    """Finds the element in the collection with the minimal distance to the given element.

    Ties are broken using the indices of the collection, earlier elements will be preferred.

    Args:
        element: The element to compare against
        collection: The collection with candidate values

    Returns: The index of the found element

    """
    distance = math.inf
    index = 0

    for i, candidate in enumerate(collection):
        candidate_distance = abs(candidate - element)
        if candidate_distance < distance:
            distance = candidate_distance
            index = i
            if distance == 0:
                return index

    return index


def get_note_durations(upper_bound_multiplier: int, lower_bound_divisor: int, base_value: int = PPQN) -> [int]:
    """Generates an array of valid note durations in ticks with regard to the PPQN.

    Automatically generates the duration of notes in ticks using the given parameters, by multiplying or dividing the
    given base value with the upper and lower bounds. In order to generate note values up to half notes with the
    standard base value of the PPQN, an upper bound of 2 has to be selected, with which the base value will be
    multiplied. In order to generate note values down to sixteenth notes with the standard base value of the PPQN,
    a lower bound of 4 as to be selected, with which the base value will be divided by.

    Args:
        upper_bound_multiplier: Maximum multiplier of the given base value
        lower_bound_divisor: Maximum divisor of the given base value
        base_value: Amount of ticks for the base value, upon which calculation is based

    Returns: An array of note values in ticks

    """
    durations = []

    i = upper_bound_multiplier
    while i >= 1:
        durations.append(int(i * base_value))
        i /= 2

    j = 2
    while j <= lower_bound_divisor:
        durations.append(int(base_value / j))
        j *= 2

    return durations


def get_tuplet_durations(note_durations, ratio_numerator, ratio_denominator) -> [int]:
    """Generates tuplet durations from a ratio and given note durations.

    Generates tuplet values for each duration in the given array, by dividing by the denominator and multiplying with
    the numerator.

    Args:
        note_durations: A base array of valid note durations
        ratio_numerator: The numerator of the tuplet
        ratio_denominator: The denominator of the tuplet

    Returns: The generated tuplet lengths in ticks

    """
    tuplet_durations = []

    for note_duration in note_durations:
        tuplet_durations.append(int((note_duration * ratio_denominator) / ratio_numerator))

    return tuplet_durations


def get_dotted_note_durations(note_durations, dotted_iterations) -> [int]:
    """Generates dotted note durations from an initial array of note durations.

    For each duration in the given array, creates up to the specified amount of iterations dotted values. E.g.,
    if `dotted_iterations` were set to 2, durations for single and twice dotted notes would be generated from the
    given array.

    Args:
        note_durations: A base array of valid note durations
        dotted_iterations: The largest amount of dots to allow for notes

    Returns: The generated note lengths in ticks

    """
    dotted_durations = []

    for dotted_note_iteration in range(dotted_iterations):
        for note_duration in note_durations:
            candidate_duration = note_duration * (1 + (1 - 1 / (2 ** (dotted_note_iteration + 1))))
            if candidate_duration.is_integer():
                dotted_durations.append(int(candidate_duration))

    return dotted_durations


def minmax(minimum, maximum, value):
    if value < minimum:
        return minimum
    if value > maximum:
        return maximum
    return value


# From http://arachnoid.com
def regress(x, terms):
    """Calculates f(x) based on x and the coefficients given by `terms`.

    Args:
        x: Input value
        terms: Coefficients, starting with x^0

    Returns:

    """
    t = 1
    r = 0
    for c in terms:
        r += c * t
        t *= x
    return r


def simple_regression(x1, y1, x2, y2, value):
    c_1 = (y2 - y1) / (x2 - x1)
    c_0 = y1 - x1 * c_1
    return c_1 * value + c_0


def velocity_from_bin(bin_index: int) -> int:
    """Returns the velocity defined by the bin.

    Args:
        bin_index: Index of the bin to get the velocity for

    Returns: The corresponding velocity

    """
    bin_size = round(VELOCITY_MAX / VELOCITY_BINS)
    return int(min(VELOCITY_MAX, (bin_index + 1) * bin_size))



import warnings


def deprecated(message):
    def decorator(func):
        def wrapper(*args, **kwargs):
            warnings.warn(f"{func.__name__} is deprecated: {message}", category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator



from __future__ import annotations

import copy
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from matplotlib import pyplot as plt, pyplot
from matplotlib.patches import Rectangle

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.midi.midi_file import MidiFile
from scoda.midi.midi_track import MidiTrack
from scoda.misc.music_theory import Key
from scoda.misc.util import minmax, simple_regression
from scoda.sequences.absolute_sequence import AbsoluteSequence
from scoda.sequences.relative_sequence import RelativeSequence
from scoda.settings.settings import PPQN, NOTE_LOWER_BOUND, NOTE_UPPER_BOUND, VELOCITY_MAX

if TYPE_CHECKING:
    from scoda.elements.bar import Bar


class Sequence:
    """Wrapper for `scoda.sequence.absolute_sequence.AbsoluteSequence` and
    `scoda.sequence.relative_sequence.RelativeSequence`.

    This class serves as a wrapper for the two above-mentioned classes. This abstraction provides an easier
    understanding for the end-user, who does not have to concern themselves with implementational details.
    """

    # General Methods

    def __init__(self, absolute_sequence: AbsoluteSequence = None, relative_sequence: RelativeSequence = None) -> None:
        super().__init__()
        self._abs_stale = True
        self._rel_stale = True

        self._difficulty = None
        self._diff_note_amount = None
        self._diff_note_values = None
        self._diff_note_classes = None
        self._diff_concurrent_notes = None
        self._diff_key = None
        self._diff_accidentals = None
        self._diff_distances = None
        self._diff_rhythm = None
        self._diff_pattern = None

        if absolute_sequence is None:
            self._abs = AbsoluteSequence()
        else:
            self._abs = absolute_sequence
            self._abs_stale = False

        if relative_sequence is None:
            self._rel = RelativeSequence()
        else:
            self._rel = relative_sequence
            self._rel_stale = False

    def __copy__(self) -> Sequence:
        copied_absolute_sequence = None if self.abs is None else copy.copy(self.abs)
        copied_relative_sequence = None if self.rel is None else copy.copy(self.rel)

        cpy = Sequence(copied_absolute_sequence, copied_relative_sequence)

        cpy._difficulty = self._difficulty
        cpy._diff_note_amount = self._diff_note_amount
        cpy._diff_note_values = self._diff_note_values
        cpy._diff_note_classes = self._diff_note_classes
        cpy._diff_concurrent_notes = self._diff_concurrent_notes
        cpy._diff_key = self._diff_key
        cpy._diff_accidentals = self._diff_accidentals
        cpy._diff_distances = self._diff_distances
        cpy._diff_rhythm = self._diff_rhythm
        cpy._diff_pattern = self._diff_pattern

        return cpy

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Sequence):
            return False

        return self.abs.__eq__(o.abs)

    @property
    def abs(self) -> AbsoluteSequence:
        """Returns a reference to the absolute representation of this sequence.

        Note that if the sequence is modified without interacting via the interface provided by this object,
        the sequence stored in this object might corrupt.

        Returns: The stored AbsoluteSequence

        """
        if self._abs_stale:
            self._abs_stale = False
            self._abs = self._rel.to_absolute_sequence()
        return self._abs

    @property
    def rel(self) -> RelativeSequence:
        """Returns a reference to the relative representation of this sequence.

        Note that if the sequence is modified without interacting via the interface provided by this object,
        the sequence stored in this object might corrupt.

        Returns: The stored RelativeSequence

        """
        if self._rel_stale:
            self._rel_stale = False
            self._rel = self._abs.to_relative_sequence()
        return self._rel

    # Basic Methods

    def add_absolute_message(self, msg) -> None:
        """See `scoda.sequence.absolute_sequence.AbsoluteSequence.add_message`."""
        self.abs.add_message(msg)
        self._rel_stale = True

    def add_relative_message(self, msg) -> None:
        """See `scoda.sequence.relative_sequence.RelativeSequence.add_message`."""
        self.rel.add_message(msg)
        self._abs_stale = True

    def concatenate(self, sequences: list[Sequence]) -> None:
        """See `scoda.sequence.relative_sequence.RelativeSequence.concatenate`."""
        self.rel.concatenate([seq.rel for seq in sequences])
        self._abs_stale = True

    def cutoff(self, maximum_length, reduced_length) -> None:
        """See `scoda.sequence.relative_sequence.AbsoluteSequence.cutoff`."""
        self.abs.cutoff(maximum_length=maximum_length, reduced_length=reduced_length)
        self._rel_stale = True

    def merge(self, sequences: list[Sequence]) -> None:
        """See `scoda.sequence.absolute_sequence.AbsoluteSequence.merge`."""
        self.abs.merge([seq.abs for seq in sequences])
        self._rel_stale = True
        self.normalise()

    def normalise(self) -> None:
        """See `scoda.sequence.relative_sequence.RelativeSequence.normalise_relative`."""
        self.rel.normalise_relative()
        self._abs_stale = True

    def pad(self, padding_length) -> None:
        """See `scoda.sequence.relative_sequence.RelativeSequence.pad`."""
        self.rel.pad(padding_length)
        self._abs_stale = True

    def save(self, file_path: str) -> MidiFile:
        """Saves the given sequence as a MIDI file.

        Args:
            file_path: Where to save the sequence to

        Returns: The resulting MidiFile

        """
        return Sequence.sequences_save([self], file_path)

    def split(self, capacities: list[int]) -> list[Sequence]:
        """See `scoda.sequence.relative_sequence.RelativeSequence.split`."""
        relative_sequences = self.rel.split(capacities)
        sequences = [Sequence(relative_sequence=seq) for seq in relative_sequences]
        return sequences

    def scale(self, factor, meta_sequence=None, quantise_afterwards=True) -> None:
        """See `scoda.sequence.relative_sequence.RelativeSequence.scale`."""
        self.rel.scale(factor, meta_sequence)
        self._abs_stale = True

        if quantise_afterwards:
            self.quantise_and_normalise()

    def transpose(self, transpose_by: int) -> bool:
        """See `scoda.sequence.relative_sequence.RelativeSequence.transpose`."""
        self._abs_stale = True
        shifted = self.rel.transpose(transpose_by)

        # Possible that notes overlap
        if shifted:
            self.normalise()
            self.quantise_note_lengths()
            self._diff_pattern = None

        if transpose_by % 12 != 0:
            self._diff_key = None
            self._diff_accidentals = None

        return shifted

    def quantise(self, step_sizes: list[int] = None) -> None:
        """See `scoda.sequence.absolute_sequence.AbsoluteSequence.quantise`."""
        self.abs.quantise(step_sizes)
        self._rel_stale = True

    def quantise_note_lengths(self, possible_durations=None, standard_length=PPQN, do_not_extend=False) -> None:
        """See `scoda.sequence.absolute_sequence.AbsoluteSequence.quantise_note_lengths`."""
        self.abs.quantise_note_lengths(possible_durations, standard_length=standard_length, do_not_extend=do_not_extend)
        self._rel_stale = True

    def quantise_and_normalise(self, step_sizes: list[int] = None, possible_durations=None, standard_length=PPQN,
                               do_not_extend=False):
        """Quantises and normalises the sequence."""
        self.quantise(step_sizes)
        self.quantise_note_lengths(possible_durations, standard_length, do_not_extend)
        self.normalise()

    # Misc. Methods

    def get_message_timings_of_type(self, message_types: [MessageType]) -> list[tuple[int, Message]]:
        """See `scoda.sequence.absolute_sequence.AbsoluteSequence.get_message_timings_of_type`.

        """
        return self.abs.get_message_timings_of_type(message_types)

    def get_sequence_duration(self) -> float:
        """See `scoda.sequence.absolute_sequence.AbsoluteSequence.get_sequence_length`.

        """
        return self.abs.get_sequence_duration()

    def get_sequence_duration_relation(self) -> float:
        """See `scoda.sequence.relative_sequence.RelativeSequence.sequence_length_relation`.

        """
        return self.rel.get_sequence_duration_relation()

    def is_empty(self) -> bool:
        """See `scoda.sequence.relative_sequence.RelativeSequence.is_empty`.

        """
        return self.rel.is_empty()

    def to_midi_track(self) -> MidiTrack:
        """See `scoda.sequence.relative_sequence.RelativeSequence.to_midi_track`."""
        return self.rel.to_midi_track()

    # Difficulty Methods

    def difficulty(self, key_signature: Key = None) -> float:
        # If difficulty not stale
        if None not in [self._difficulty, self._diff_note_amount, self._diff_note_values, self._diff_note_classes,
                        self._diff_key, self._diff_accidentals, self._diff_distances, self._diff_rhythm,
                        self._diff_pattern, self._diff_concurrent_notes]:
            return self._difficulty

        self.normalise()

        if key_signature is None:
            key_signature = self.rel.get_key_signature_guess()

        difficulty_weights = [
            (self.diff_note_values, 0, 0.45),
            (self.diff_note_amount, 0, 0.5),
            (self.diff_concurrent_notes, 0, 0.475),
            (self.diff_distances, 0, 0.15),
            (self.diff_rhythm, -0.1, 0.2),
            (self.diff_key(key_signature), -0.05, 0.15),
            (self.diff_accidentals(key_signature), -0.05, 0.1),
            (self.diff_note_classes, -0.1, 0.15),
        ]

        difficulty_percentages = [
            (self.diff_pattern, -0.4, 0),
        ]

        overall_difficulty = 0

        for difficulty_weight in difficulty_weights:
            change = simple_regression(0, difficulty_weight[1], 1, difficulty_weight[2], difficulty_weight[0])
            overall_difficulty += change

        percentage_change = 0
        for percentage_weight in difficulty_percentages:
            change = simple_regression(0, percentage_weight[1], 1, percentage_weight[2], percentage_weight[0])
            percentage_change += change
        overall_difficulty *= (1 + percentage_change)

        overall_difficulty = minmax(0, 1, overall_difficulty)

        self._difficulty = overall_difficulty
        return self._difficulty

    @property
    def diff_note_amount(self) -> float:
        if self._diff_note_amount is None:
            self._diff_note_amount = self.rel.diff_note_amount()
        return self._diff_note_amount

    @property
    def diff_note_values(self) -> float:
        if self._diff_note_values is None:
            self._diff_note_values = self.abs.diff_note_values()
        return self._diff_note_values

    @property
    def diff_note_classes(self) -> float:
        if self._diff_note_classes is None:
            self._diff_note_classes = self.rel.diff_note_classes()
        return self._diff_note_classes

    @property
    def diff_concurrent_notes(self) -> float:
        if self._diff_concurrent_notes is None:
            self._diff_concurrent_notes = self.rel.diff_concurrent_notes()
        return self._diff_concurrent_notes

    def diff_key(self, key_signature) -> float:
        if self._diff_key is None:
            self._diff_key = self.rel.diff_key(key=key_signature)
        return self._diff_key

    def diff_accidentals(self, key_signature) -> float:
        if self._diff_accidentals is None:
            self._diff_accidentals = self.rel.diff_accidentals(key=key_signature)
        return self._diff_accidentals

    @property
    def diff_distances(self) -> float:
        if self._diff_distances is None:
            self._diff_distances = self.rel.diff_distances()
        return self._diff_distances

    @property
    def diff_rhythm(self) -> float:
        if self._diff_rhythm is None:
            self._diff_rhythm = self.abs.diff_rhythm()
        return self._diff_rhythm

    @property
    def diff_pattern(self) -> float:
        if self._diff_pattern is None:
            self._diff_pattern = self.rel.diff_pattern()
        return self._diff_pattern

    # Static Functions

    @staticmethod
    def sequences_load(file_path: Path | str = None,
                       midi_file: MidiFile = None,
                       track_indices: list[list[int]] = None,
                       meta_track_indices: list[int] = None,
                       target_meta_track_index: int = 0) -> list[Sequence]:
        """Creates `scoda.Sequence` objects from the provided MIDI file.

        Args:
            midi_file: If provided, this file is used instead of trying to load from a file.
            file_path: The file path of the MIDI file.
            track_indices: A list of lists indicating which tracks of the MIDI file should be merged into which tracks
                of the resulting sequence.
            meta_track_indices: A list of indices of tracks of the MIDI file to consider for meta messages.
            target_meta_track_index: The index of the track of the final sequence that should contain meta messages.

        Returns: A list of `scoda.Sequence` objects.

        """
        if midi_file is None:
            midi_file = MidiFile.open(file_path)

        if track_indices is None:
            track_indices = [[i] for i, _ in enumerate(midi_file.tracks)]
        if meta_track_indices is None:
            meta_track_indices = [i for i, _ in enumerate(midi_file.tracks)]

        # Get sequence from MIDI file
        merged_sequences = midi_file.convert(track_indices, meta_track_indices,
                                             meta_track_index=target_meta_track_index)

        return merged_sequences

    @staticmethod
    def sequences_save(sequences: list[Sequence],
                       file_path: Path | str) -> MidiFile:
        """Saves the given sequence as a MIDI file.

        Args:
            sequences: Sequences to save
            file_path: Where to save the sequence to

        Returns: The resulting MidiFile

        """
        midi_file = MidiFile()
        for sequence in sequences:
            midi_file.tracks.append(sequence.to_midi_track())
        midi_file.save(file_path)

        return midi_file

    @staticmethod
    def sequences_split_bars(sequences_input: list[Sequence],
                             meta_track_index=0,
                             quantise_note_lengths=True) -> list[list[Bar]]:
        """Splits the sequence into a lists of `scoda.Bar`, conforming to the contained time signatures.

        Each list of bars will correspond to one of the given sequence.

        Args:
            sequences_input: The sequence to split.
            meta_track_index: The index of the sequence that contains the time signature changes.
            quantise_note_lengths: Whether the note lengths of the split bar should be quantised after the split.

        Returns: A list of bars.

        """
        from scoda.elements.bar import Bar

        sequences = copy.copy(sequences_input)

        # Split into bars, carry key and time signature
        current_point_in_time = 0
        current_ts_numerator = 4
        current_ts_denominator = 4
        current_key = None

        # Determine signature timings
        meta_track = sequences[meta_track_index]
        time_signature_timings = meta_track.get_message_timings_of_type([MessageType.TIME_SIGNATURE])
        key_signature_timings = meta_track.get_message_timings_of_type([MessageType.KEY_SIGNATURE])

        tracks_bars = [[] for _ in sequences]

        if len(time_signature_timings) == 0:
            time_signature_timings = [(0,
                                       Message(message_type=MessageType.TIME_SIGNATURE, numerator=current_ts_numerator,
                                               denominator=current_ts_denominator))]

        # Keep track of when bars are of equal length
        tracks_synchronised = False

        # Repeat until all tracks of exactly equal length
        while not tracks_synchronised:
            # Obtain new time or key signatures
            time_signature = next((timing for timing in time_signature_timings if timing[0] <= current_point_in_time)
                                  , None)
            key_signature = next((timing for timing in key_signature_timings if timing[0] <= current_point_in_time)
                                 , None)

            # Update time signature
            if time_signature is not None:
                time_signature_timings.pop(0)
                current_ts_numerator = time_signature[1].numerator
                current_ts_denominator = time_signature[1].denominator

            # Update key signature
            if key_signature is not None:
                key_signature_timings.pop(0)
                current_key = key_signature[1].key

            # Calculate length of current bar based on time signature
            length_bar = int(PPQN * (current_ts_numerator / (current_ts_denominator / 4)))
            current_point_in_time += length_bar

            # Assume after this split, tracks are synchronized
            tracks_synchronised = True

            # Split sequence into bars
            for i, sequence in enumerate(sequences):
                split_up = sequence.split([length_bar])

                # Check if we reached the end of the sequence
                if len(split_up) > 1:
                    tracks_synchronised = False
                    sequences[i] = split_up[1]
                # Fill with placeholder empty sequence
                else:
                    if len(split_up) == 0:
                        split_up.append(Sequence())
                    sequences[i] = Sequence()

                # Quantise note lengths, in case splitting into bars affected them
                sequence_to_add = split_up[0]
                if quantise_note_lengths:
                    sequence_to_add.quantise_note_lengths(do_not_extend=True)

                # Append split bar to list of bars
                tracks_bars[i].append(
                    Bar(sequence_to_add, current_ts_numerator, current_ts_denominator,
                        Key(current_key) if current_key is not None else None))

        return tracks_bars

    @staticmethod
    def plot_pianorolls(sequences: [Sequence],
                        title: str = None,
                        x_label: str = None,
                        y_label: str = None,
                        x_scale: list[int] = None,
                        y_scale: list[int] = (NOTE_LOWER_BOUND, NOTE_UPPER_BOUND + 1),
                        show_velocity: bool = True,
                        x_tick_spacing=PPQN) -> pyplot:
        """Creates a piano roll from the given sequence.

        Creates a visualisation in form of a piano roll from the given sequence, where each note is visualised using
        a rectangle. The opacity of the rectangles corresponds to the velocity the notes are played with. Options for
        scaling the representation are given, where in the default case height and width of the piano roll are
        selected in such a way that all notes fit exactly.

        Args:
            sequences: The sequence to create a representation for
            title: An optional title to set
            x_label: Label of the x-axis
            y_label: Label of the y-axis
            x_scale: The scale of the x-axis, if not given will be chosen in such a way that all notes fit exactly
            y_scale: The scale of the y-axis, if not given will be chosen in such a way that all notes fit exactly
            show_velocity: Whether to show the velocity by changing the opacity of some notes
            x_tick_spacing: Spacing of the ticks on the x-axis

        """
        # Create new figure
        fig = plt.figure(dpi=300)

        # Create subplots for each of the sequence
        gs = fig.add_gridspec(len(sequences), hspace=0.1)
        axs = gs.subplots(sharex=True, sharey=True)

        # Keep track of length and range of sequence
        x_scale_max = 0
        y_scale_min = NOTE_UPPER_BOUND
        y_scale_max = NOTE_LOWER_BOUND

        # Workaround for single sequence
        if len(sequences) == 1:
            axs = [axs]

        # Draw notes
        for i, sequence in enumerate(sequences):
            note_array = sequence.abs.get_message_time_pairings()

            for note in note_array:
                start_time = note[0].time
                duration = note[1].time - start_time
                pitch = note[0].note

                # Keep track of scales
                if pitch < y_scale_min:
                    y_scale_min = pitch
                if pitch > y_scale_max:
                    y_scale_max = pitch

                # Calculate opacity based on velocity
                opacity = simple_regression(1, 1, 0, 0.5, note[0].velocity / VELOCITY_MAX)

                # Draw rectangle
                axs[i].add_patch(
                    Rectangle((start_time, pitch), duration, 1,
                              facecolor=(0, 0, 0, 1 if not show_velocity else opacity)))

            # Get length of sequence (if wait messages occur after notes)
            length = 0
            for msg in sequence.rel.messages:
                if msg.message_type == MessageType.WAIT:
                    length += msg.time

            x_scale_max = max(x_scale_max, length)

        # Define scale of plot
        if x_scale is None:
            x_scale = [0, x_scale_max]
        if y_scale is None:
            y_scale = [y_scale_min, y_scale_max + 1]
        else:
            y_scale = [y_scale[0], y_scale[1]]

        for ax in axs:
            ax.label_outer()

            # X Ticks
            ax.set_xticks(ticks=np.arange(0, ((x_scale[1] / x_tick_spacing) + 1) * x_tick_spacing, x_tick_spacing))

            # Y Ticks
            ax.set_yticks(ticks=np.arange(24, 24 + (8 + 1) * 12, 12),
                          labels=["C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "C9"])
            ax.set_yticks(ticks=np.arange(NOTE_LOWER_BOUND, NOTE_UPPER_BOUND + 1, 1), minor=True)

            # Activate grid
            ax.grid(visible=True)

        # Title plot
        plt.suptitle(title, fontsize=20)

        # Label axis
        fig.supxlabel(x_label)
        fig.supylabel(y_label)

        # Define scale
        plt.xlim(x_scale)
        plt.ylim(y_scale)

        return plt

    # Private Functions

    @staticmethod
    def _fill_dictionary_entry(entry,
                               msg_type=None,
                               time=None,
                               note=None,
                               velocity=None,
                               control=None,
                               program=None,
                               numerator=None,
                               denominator=None,
                               key=None,
                               rel_note_dist=None,
                               rel_note_pair_dist=None,
                               rel_note_pair_oct=None):
        entry["msg_type"] = msg_type
        entry["time"] = time
        entry["note"] = note
        entry["velocity"] = velocity
        entry["control"] = control
        entry["program"] = program
        entry["numerator"] = numerator
        entry["denominator"] = denominator
        entry["key"] = key
        entry["rel_distance"] = rel_note_dist
        entry["rel_note_pair_dist"] = rel_note_pair_dist
        entry["rel_note_pair_oct"] = rel_note_pair_oct



import math
from abc import ABC, abstractmethod
from typing import Any

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.enumerations.tokenisation_flags import TokenisationFlags
from scoda.exceptions.tokenisation_exception import TokenisationException
from scoda.misc.music_theory import CircleOfFifths
from scoda.misc.scoda_logging import get_logger
from scoda.sequences.sequence import Sequence
from scoda.settings.settings import PPQN
from scoda.tokenisation.base_tokenisation import BaseTokeniser

LOGGER = get_logger(__name__)


class BaseNotelikeTokeniser(BaseTokeniser, ABC):

    def __init__(self, running_value: bool, running_time_sig: bool) -> None:
        super().__init__()

        self.flags[TokenisationFlags.RUNNING_VALUE] = running_value
        self.flags[TokenisationFlags.RUNNING_TIME_SIG] = running_time_sig


class BaseLargeVocabularyNotelikeTokeniser(BaseNotelikeTokeniser, ABC):
    SUPPORTED_VALUES = [2, 3, 4, 6, 8, 9, 12, 16, 18, 24, 32, 36, 48, 64, 72, 96]
    NOTE_SECTION_SIZE = None

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__(False, running_time_sig)

    def tokenise(self, sequence: Sequence, apply_buffer: bool = True, reset_time: bool = True) -> list[int]:
        tokens = []
        event_pairings = sequence.abs.get_message_time_pairings(
            [MessageType.NOTE_ON, MessageType.NOTE_OFF, MessageType.TIME_SIGNATURE, MessageType.INTERNAL])

        for event_pairing in event_pairings:
            msg_type = event_pairing[0].message_type
            msg_time = event_pairing[0].time

            if msg_type == MessageType.NOTE_ON:
                msg_note = event_pairing[0].note
                msg_value = event_pairing[1].time - msg_time

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")
                if msg_value not in LargeVocabularyNotelikeTokeniser.SUPPORTED_VALUES:
                    raise TokenisationException(f"Invalid note value: {msg_value}")

                # Check if message occurs at current time, if not place rest messages
                if not self.cur_time == msg_time:
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                    self.cur_time += self.cur_rest_buffer
                    self.cur_rest_buffer = 0

                # Callback
                self._tokenise_note(tokens, msg_note, msg_value)

                self.cur_time_target = max(self.cur_time_target, self.cur_time + msg_value)
                self.prv_note = msg_note
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = event_pairing[0].numerator
                msg_denominator = event_pairing[0].denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                # Check if time signature has to be defined
                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                    self.cur_time += self.cur_rest_buffer
                    self.cur_rest_buffer = 0

                    tokens.append(numerator - 2 + len(self.SUPPORTED_VALUES) * self.NOTE_SECTION_SIZE + 4 + 24)

                self.prv_numerator = numerator
            elif msg_type == MessageType.INTERNAL:
                self.cur_rest_buffer += msg_time - self.cur_time

        if apply_buffer:
            self.cur_rest_buffer = max(self.cur_time_target - self.cur_time, self.cur_rest_buffer)
            tokens.extend(
                self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
            self.cur_time += self.cur_rest_buffer
            self.cur_rest_buffer = 0

        if reset_time:
            self.reset_time()

        return tokens

    @abstractmethod
    def _tokenise_note(self, tokens: list[int], msg_note: int, msg_value: int) -> None:
        pass


class StandardNotelikeTokeniser(BaseNotelikeTokeniser):
    """Tokeniser that uses note-like temporal representation.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [        4] ... wait
    [  5 -  28] ... value definition
    [ 29 - 116] ... note
    [117 - 131] ... time signature numerator in eights from 2/8 to 16/8
    [      132] ... note with previous pitch (running pitch only)
    """

    VOCAB_SIZE = 133

    def __init__(self, running_value: bool, running_pitch: bool, running_time_sig: bool) -> None:
        super().__init__(running_value, running_time_sig)

        self.flags[TokenisationFlags.RUNNING_PITCH] = running_pitch

    def tokenise(self, sequence: Sequence, apply_buffer: bool = True, reset_time: bool = True) -> list[int]:
        tokens = []
        event_pairings = sequence.abs.get_message_time_pairings(
            [MessageType.NOTE_ON, MessageType.NOTE_OFF, MessageType.TIME_SIGNATURE, MessageType.INTERNAL])

        for event_pairing in event_pairings:
            msg_type = event_pairing[0].message_type
            msg_time = event_pairing[0].time

            if msg_type == MessageType.NOTE_ON:
                msg_note = event_pairing[0].note
                msg_value = event_pairing[1].time - msg_time

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                # Check if message occurs at current time, if not place rest messages
                if not self.cur_time == msg_time:
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4, index_time_def=5))

                # Check if value of note has to be defined
                if not (self.prv_value == msg_value and self.flags.get(TokenisationFlags.RUNNING_VALUE, False)):
                    tokens.extend(self._general_tokenise_flush_time_buffer(msg_value, index_time_def=5))

                # Check if pitch of note has to be defined
                if not (self.prv_note == msg_note and self.flags.get(TokenisationFlags.RUNNING_PITCH, False)):
                    tokens.append(msg_note - 21 + 29)
                else:
                    tokens.append(132)

                self.cur_time_target = max(self.cur_time_target, self.cur_time + msg_value)
                self.prv_type = MessageType.NOTE_ON
                self.prv_note = msg_note
                self.prv_value = msg_value
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = event_pairing[0].numerator
                msg_denominator = event_pairing[0].denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                # Check if time signature has to be defined
                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4, index_time_def=5))
                    tokens.append(numerator - 2 + 117)

                self.prv_type = MessageType.TIME_SIGNATURE
                self.prv_numerator = numerator
            elif msg_type == MessageType.INTERNAL:
                self.cur_rest_buffer += msg_time - self.cur_time
                self.prv_type = MessageType.INTERNAL

        if apply_buffer:
            tokens.extend(
                self._notelike_tokenise_flush_rest_buffer(apply_target=True, wait_token=4, index_time_def=5))

        if reset_time:
            self.reset_time()

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        cur_time = 0
        prv_type = None
        prv_note = None
        prv_value = math.nan

        for token in tokens:
            if token <= 2:
                prv_type = "sequence_control"
            elif token == 3:
                prv_type = "sequence_control"
            elif token == 4:
                cur_time += prv_value
                prv_type = MessageType.WAIT
            elif 5 <= token <= 28:
                if prv_type == MessageType.INTERNAL:
                    prv_value += token - 4
                else:
                    prv_value = token - 4
                prv_type = MessageType.INTERNAL
            elif 29 <= token <= 116:
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_ON, note=token - 29 + 21, time=cur_time))
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_OFF, note=token - 29 + 21,
                            time=cur_time + prv_value))
                prv_type = MessageType.NOTE_ON
                prv_note = token - 29 + 21
            elif 117 <= token <= 131:
                seq.add_absolute_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, time=cur_time,
                            numerator=token - 117 + 2, denominator=8)
                )
                prv_type = MessageType.TIME_SIGNATURE
            elif token == 132:
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_ON, note=prv_note, time=cur_time))
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_OFF, note=prv_note, time=cur_time + prv_value))
                prv_type = MessageType.NOTE_ON
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq

    def get_constraints(self, tokens: list[int], previous_state: dict = None, min_bars: int = -1) -> \
            tuple[list[int], dict[str, Any]]:
        running_value = self.flags.get(TokenisationFlags.RUNNING_VALUE, False)
        running_pitch = self.flags.get(TokenisationFlags.RUNNING_PITCH, False)
        running_time_sig = self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)

        cur_time = 0
        cur_bar = 0
        bar_time = 0
        bar_tokens = []
        bar_capacity = None
        pnt_concurrent_notes = []
        seq_started = False
        seq_stopped = False
        prv_type = None
        prv_note = None
        prv_value = math.nan

        if previous_state is not None:
            cur_time = previous_state["cur_time"]
            cur_bar = previous_state["cur_bar"]
            bar_time = previous_state["bar_time"]
            bar_tokens = previous_state["bar_tokens"]
            pnt_concurrent_notes = previous_state["pnt_concurrent_notes"]
            bar_capacity = previous_state["bar_capacity"]
            seq_started = previous_state["seq_started"]
            seq_stopped = previous_state["seq_stopped"]
            prv_type = previous_state["prv_type"]
            prv_note = previous_state["prv_note"]
            prv_value = previous_state["prv_value"]

        # === Retrace sequence ===

        for token in tokens:
            if token == 0:
                bar_tokens.append(token)
                prv_type = MessageType.SEQUENCE_CONTROL
            elif token == 1:
                bar_tokens.append(token)
                seq_started = True
                prv_type = MessageType.SEQUENCE_CONTROL
            elif token == 2:
                bar_tokens.append(token)
                seq_stopped = True
                prv_type = MessageType.SEQUENCE_CONTROL
            elif token == 3:
                cur_bar += 1
                bar_time = 0
                bar_tokens.append(token)
                prv_type = MessageType.SEQUENCE_CONTROL
            elif token == 4:
                cur_time += prv_value
                bar_time += prv_value

                while bar_time > bar_capacity:
                    bar_time -= bar_capacity
                    cur_bar += 1

                bar_tokens.append(token)
                pnt_concurrent_notes = []
                prv_type = MessageType.WAIT

            elif 5 <= token <= 28:
                bar_tokens.append(token)
                if prv_type != MessageType.INTERNAL:
                    prv_value = 0
                prv_value += token - 4
                prv_type = MessageType.INTERNAL
            elif 29 <= token <= 116:
                bar_tokens.append(token)
                pnt_concurrent_notes.append(token - 29 + 21)
                prv_type = MessageType.NOTE_ON
                prv_note = token - 29 + 21
            elif 117 <= token <= 131:
                bar_tokens.append(token)
                bar_capacity = int(((token - 117 + 2) / 8) * 4 * PPQN)
                prv_type = MessageType.TIME_SIGNATURE
            elif token == 132:
                bar_tokens.append(token)
                pnt_concurrent_notes.append(prv_note)
                prv_type = MessageType.NOTE_ON
            else:
                raise TokenisationException(f"Encountered invalid token during constraint creation: {token}")

            if not running_value and not 5 <= token <= 28:
                prv_value = None

        # === Retrieve valid tokens ===

        # Helper variables
        h_started_not_stopped = seq_started and not seq_stopped
        h_bar_contains_time_signature = any(117 <= t <= 131 for t in bar_tokens)
        h_bar_valid_time_signature = (running_time_sig and bar_capacity is not None) or h_bar_contains_time_signature
        h_note_not_open = lambda x: x is not None and x not in pnt_concurrent_notes
        h_has_prv_value = prv_value is not None

        valid_tokens = []

        # [        0] ... pad
        if seq_started and seq_stopped:
            valid_tokens.append(0)
        # [        1] ... start
        if not seq_started:
            valid_tokens.append(1)
        # [        2] ... stop
        if not seq_stopped and (min_bars == -1 or cur_bar + 1 == min_bars) and bar_time == bar_capacity:
            valid_tokens.append(2)
        # [        3] ... bar separator
        if h_started_not_stopped and h_bar_valid_time_signature and bar_time == bar_capacity and 3 not in bar_tokens:
            valid_tokens.append(3)
        # [        4] ... wait
        if h_started_not_stopped and h_bar_valid_time_signature and h_has_prv_value and bar_capacity - bar_time >= prv_value:
            valid_tokens.append(4)
        # [  5 -  28] ... value definition
        if h_started_not_stopped and h_bar_valid_time_signature:
            prv_value_volatile = prv_value
            if prv_type != MessageType.INTERNAL:
                prv_value_volatile = 0
            for v_d in range(1, 24 + 1):
                if bar_capacity - bar_time >= prv_value_volatile + v_d:
                    valid_tokens.append(v_d + 4)
        # [ 29 - 116] ... note
        if h_started_not_stopped and h_bar_valid_time_signature and h_has_prv_value:
            for n in range(21, 108 + 1):
                if n not in pnt_concurrent_notes:
                    valid_tokens.append(n - 21 + 29)
        # [117 - 131] ... time signature numerator in eights from 2/8 to 16/8
        if h_started_not_stopped and not h_bar_contains_time_signature and bar_time == 0:
            for t in range(117, 131 + 1):
                valid_tokens.append(t)
        # [      132] ... note with previous pitch (running pitch only)
        if h_started_not_stopped and h_bar_valid_time_signature and running_pitch and h_has_prv_value and h_note_not_open(
                prv_note):
            valid_tokens.append(132)

        # === Store and return finding ===

        state = {"cur_time": cur_time,
                 "cur_bar": cur_bar,
                 "bar_time": bar_time,
                 "bar_tokens": bar_tokens,
                 "pnt_concurrent_notes": pnt_concurrent_notes,
                 "bar_capacity": bar_capacity,
                 "seq_started": seq_started,
                 "seq_stopped": seq_stopped,
                 "prv_type": prv_type,
                 "prv_note": prv_note,
                 "prv_value": prv_value}

        return valid_tokens, state


class LargeVocabularyNotelikeTokeniser(BaseLargeVocabularyNotelikeTokeniser):
    """Tokeniser that uses note-like temporal representation.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [  4 -  27] ... wait
    [ 28 - 115] ... notes with duration of 2 ticks
    [116 - 203] ... notes with duration of 3 ticks
    [204 - 291] ... notes with duration of 4 ticks
    [292 - 379] ... notes with duration of 6 ticks
    [380 - 467] ... notes with duration of 8 ticks
    [468 - 555] ... notes with duration of 9 ticks
    [556 - 643] ... notes with duration of 12 ticks
    [644 - 731] ... notes with duration of 16 ticks
    [732 - 819] ... notes with duration of 18 ticks
    [820 - 907] ... notes with duration of 24 ticks
    [908 - 995] ... notes with duration of 32 ticks
    [996 -1083] ... notes with duration of 36 ticks
    [1084-1171] ... notes with duration of 48 ticks
    [1172-1259] ... notes with duration of 64 ticks
    [1260-1347] ... notes with duration of 72 ticks
    [1348-1435] ... notes with duration of 96 ticks
    [1436-1450] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 1451

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__(running_time_sig)

    NOTE_SECTION_SIZE = 88

    def _tokenise_note(self, tokens: list[int], msg_note: int, msg_value: int) -> None:
        # Add token representing pitch and value
        tokens.append(
            msg_note - 21 + 28 + LargeVocabularyNotelikeTokeniser.SUPPORTED_VALUES.index(msg_value) * 88)

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        cur_time = 0
        note_section_size = LargeVocabularyNotelikeTokeniser.NOTE_SECTION_SIZE

        boundary_token_ts = len(LargeVocabularyNotelikeTokeniser.SUPPORTED_VALUES) * note_section_size + 4 + 24

        for token in tokens:
            if token <= 3:
                pass
            elif 4 <= token <= 27:
                cur_time += token - 3
            elif 28 <= token <= boundary_token_ts - 1:
                note_pitch = (token - 28) % note_section_size + 21
                note_value = LargeVocabularyNotelikeTokeniser.SUPPORTED_VALUES[(token - 28) // note_section_size]

                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_ON, note=note_pitch, time=cur_time))
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_OFF, note=note_pitch, time=cur_time + note_value))
            elif boundary_token_ts <= token <= boundary_token_ts + 14:
                seq.add_absolute_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, time=cur_time,
                            numerator=token - boundary_token_ts + 2,
                            denominator=8)
                )
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq


class RelativeNotelikeTokeniser(BaseNotelikeTokeniser):
    """Tokeniser that uses note-like temporal representation with relative distances between notes.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [        4] ... wait
    [  5 -  28] ... value definition
    [ 29 - 203] ... note with relative distance to previous note
    [204 - 219] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 220

    def __init__(self, running_value: bool, running_time_sig: bool) -> None:
        super().__init__(running_value, running_time_sig)

        self.reset()

    def reset_previous(self) -> None:
        super().reset_previous()

        # A4 as base note
        self.prv_note = 69

    def tokenise(self, sequence: Sequence, apply_buffer: bool = True, reset_base_note: bool = True,
                 reset_time: bool = True) -> list[int]:
        tokens = []
        event_pairings = sequence.abs.get_message_time_pairings(
            [MessageType.NOTE_ON, MessageType.NOTE_OFF, MessageType.TIME_SIGNATURE, MessageType.INTERNAL])

        for event_pairing in event_pairings:
            msg_type = event_pairing[0].message_type
            msg_time = event_pairing[0].time

            if msg_type == MessageType.NOTE_ON:
                msg_note = event_pairing[0].note
                msg_value = event_pairing[1].time - msg_time

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                # Check if message occurs at current time, if not place rest messages
                if not self.cur_time == msg_time:
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4, index_time_def=5))

                # Check if value of note has to be defined
                if not (self.prv_value == msg_value and self.flags.get(TokenisationFlags.RUNNING_VALUE, False)):
                    tokens.extend(self._general_tokenise_flush_time_buffer(msg_value, index_time_def=5))

                # Get distance
                note_distance = msg_note - self.prv_note
                tokens.append(note_distance + 87 + 29)

                self.cur_time_target = max(self.cur_time_target, self.cur_time + msg_value)
                self.prv_type = MessageType.NOTE_ON
                self.prv_note = msg_note
                self.prv_value = msg_value
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = event_pairing[0].numerator
                msg_denominator = event_pairing[0].denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                # Check if time signature has to be defined
                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4, index_time_def=5))
                    tokens.append(numerator - 2 + 204)

                self.prv_type = MessageType.TIME_SIGNATURE
                self.prv_numerator = numerator
            elif msg_type == MessageType.INTERNAL:
                self.cur_rest_buffer += msg_time - self.cur_time
                self.prv_type = MessageType.INTERNAL

        if apply_buffer:
            tokens.extend(
                self._notelike_tokenise_flush_rest_buffer(apply_target=True, wait_token=4, index_time_def=5))

        if reset_time:
            self.reset_time()

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        cur_time = 0
        prv_type = None
        prv_note = 69  # A4 is base note
        prv_value = math.nan

        for token in tokens:
            if token <= 2:
                prv_type = "sequence_control"
            elif token == 3:
                prv_type = "sequence_control"
            elif token == 4:
                cur_time += prv_value
                prv_type = MessageType.WAIT
            elif 5 <= token <= 28:
                if prv_type == MessageType.INTERNAL:
                    prv_value += token - 4
                else:
                    prv_value = token - 4
                prv_type = MessageType.INTERNAL
            elif 29 <= token <= 203:
                note = prv_note + token - 29 - 87
                prv_note = note

                if not (21 <= note <= 108):
                    LOGGER.warning(f"Invalid note: {note}")
                while note < 21:
                    note += 12
                while note > 108:
                    note -= 12

                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_ON, note=note, time=cur_time))
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_OFF, note=note, time=cur_time + prv_value))
                prv_type = MessageType.NOTE_ON
            elif 204 <= token <= 219:
                seq.add_absolute_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, time=cur_time,
                            numerator=token - 204 + 2, denominator=8)
                )
                prv_type = MessageType.TIME_SIGNATURE
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq


class CoFNotelikeTokeniser(BaseNotelikeTokeniser):
    """Tokeniser that uses note-like temporal representation with circle of fifths distances between notes.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [        4] ... wait
    [  5 -  28] ... value definition
    [ 29 -  45] ... octave shift between notes
    [ 46 -  57] ... note without octave in distance on the circle of fifths
    [ 58 -  73] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 74

    def __init__(self, running_value: bool, running_octave: bool, running_time_sig: bool) -> None:
        super().__init__(running_value, running_time_sig)

        self.flags[TokenisationFlags.RUNNING_OCTAVE] = running_octave

        self.prv_octave = None

        self.reset()

    def reset_previous(self) -> None:
        super().reset_previous()

        # A4 as base note
        self.prv_note = 69
        self.prv_octave = 4

    def tokenise(self, sequence: Sequence, apply_buffer: bool = True, reset_base_note: bool = True,
                 reset_time: bool = True) -> list[int]:
        tokens = []
        event_pairings = sequence.abs.get_message_time_pairings(
            [MessageType.NOTE_ON, MessageType.NOTE_OFF, MessageType.TIME_SIGNATURE, MessageType.INTERNAL])

        for event_pairing in event_pairings:
            msg_type = event_pairing[0].message_type
            msg_time = event_pairing[0].time

            if msg_type == MessageType.NOTE_ON:
                msg_note = event_pairing[0].note
                msg_value = event_pairing[1].time - msg_time

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                # Check if message occurs at current time, if not place rest messages
                if not self.cur_time == msg_time:
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4, index_time_def=5))

                # Check if value of note has to be defined
                if not (self.prv_value == msg_value and self.flags.get(TokenisationFlags.RUNNING_VALUE, False)):
                    tokens.extend(self._general_tokenise_flush_time_buffer(msg_value, index_time_def=5))

                # Get distances
                octave_tgt = msg_note // 12 - 1
                octave_src = self.prv_note // 12 - 1
                octave_shift = octave_tgt - octave_src
                assert -8 <= octave_shift <= 8

                cof_dist = CircleOfFifths.get_distance(self.prv_note, msg_note)
                assert -5 <= cof_dist <= 6

                # Insert octave shift (if necessary) and note distance
                if not (self.prv_octave == octave_tgt and self.flags.get(TokenisationFlags.RUNNING_OCTAVE, False)):
                    tokens.append((octave_shift + 8) + 29)
                tokens.append((cof_dist + 5) + 46)

                self.cur_time_target = max(self.cur_time_target, self.cur_time + msg_value)
                self.prv_type = MessageType.NOTE_ON
                self.prv_note = msg_note
                self.prv_octave = octave_tgt
                self.prv_value = msg_value
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = event_pairing[0].numerator
                msg_denominator = event_pairing[0].denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                # Check if time signature has to be defined
                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    self.cur_rest_buffer += msg_time - self.cur_time
                    tokens.extend(
                        self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4, index_time_def=5))
                    tokens.append(numerator - 2 + 58)

                self.prv_type = MessageType.TIME_SIGNATURE
                self.prv_numerator = numerator
            elif msg_type == MessageType.INTERNAL:
                self.cur_rest_buffer += msg_time - self.cur_time
                self.prv_type = MessageType.INTERNAL

        if apply_buffer:
            tokens.extend(
                self._notelike_tokenise_flush_rest_buffer(apply_target=True, wait_token=4, index_time_def=5))

        if reset_time:
            self.reset_time()

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        cur_time = 0
        prv_type = None
        prv_note = 69  # A4 is base note
        prv_octave = 4
        prv_value = math.nan

        for token in tokens:
            if token <= 2:
                prv_type = "sequence_control"
            elif token == 3:
                prv_type = "sequence_control"
            elif token == 4:
                cur_time += prv_value
                prv_type = MessageType.WAIT
            elif 5 <= token <= 28:
                if prv_type == MessageType.INTERNAL:
                    prv_value += token - 4
                else:
                    prv_value = token - 4
                prv_type = MessageType.INTERNAL
            elif 29 <= token <= 45:
                prv_octave += token - 29 - 8
            elif 46 <= token <= 57:
                note_base = CircleOfFifths.from_distance(prv_note, (token - 46) - 5)
                note = note_base + prv_octave * 12 + 12  # Shifts notes to A0

                if not (21 <= note <= 108):
                    LOGGER.warning(f"Invalid note: {note}")
                while note < 21:
                    note += 12
                while note > 108:
                    note -= 12

                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_ON, note=note, time=cur_time))
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_OFF, note=note, time=cur_time + prv_value))
                prv_type = MessageType.NOTE_ON
                prv_note = note
            elif 58 <= token <= 73:
                seq.add_absolute_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, time=cur_time,
                            numerator=token - 58 + 2, denominator=8)
                )
                prv_type = MessageType.TIME_SIGNATURE
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq


class LargeVocabularyCoFNotelikeTokeniser(BaseLargeVocabularyNotelikeTokeniser):
    """Tokeniser that uses note-like temporal representation with circle of fifths distances between notes.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [  4 -  27] ... wait
    [ 28 - 231] ... notes with duration of 2 ticks
    [232 - 435] ... notes with duration of 3 ticks
    [436 - 639] ... notes with duration of 4 ticks
    [640 - 843] ... notes with duration of 6 ticks
    [844 -1047] ... notes with duration of 8 ticks
    [1048-1251] ... notes with duration of 9 ticks
    [1252-1455] ... notes with duration of 12 ticks
    [1456-1659] ... notes with duration of 16 ticks
    [1660-1863] ... notes with duration of 18 ticks
    [1864-2067] ... notes with duration of 24 ticks
    [2068-2271] ... notes with duration of 32 ticks
    [2272-2475] ... notes with duration of 36 ticks
    [2476-2679] ... notes with duration of 48 ticks
    [2680-2883] ... notes with duration of 64 ticks
    [2884-3087] ... notes with duration of 72 ticks
    [3088-3291] ... notes with duration of 96 ticks
    [3292-3307] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 3308

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__(running_time_sig)

        self.prv_octave = None

        self.reset()

    NOTE_SECTION_SIZE = 204

    def reset_previous(self) -> None:
        super().reset_previous()

        # A4 as base note
        self.prv_note = 69
        self.prv_octave = 4

    def _tokenise_note(self, tokens: list[int], msg_note: int, msg_value: int) -> None:
        # Get distances
        octave_tgt = msg_note // 12 - 1
        octave_src = self.prv_note // 12 - 1
        octave_shift = octave_tgt - octave_src
        assert -8 <= octave_shift <= 8

        cof_dist = CircleOfFifths.get_distance(self.prv_note, msg_note)
        assert -5 <= cof_dist <= 6

        token_value = (28 +
                       (cof_dist + 5) +
                       ((octave_shift + 8) * 12) +
                       self.SUPPORTED_VALUES.index(msg_value) * LargeVocabularyCoFNotelikeTokeniser.NOTE_SECTION_SIZE)

        tokens.append(token_value)

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        cur_time = 0
        prv_note = 69  # A4 is base note
        note_section_size = LargeVocabularyCoFNotelikeTokeniser.NOTE_SECTION_SIZE

        boundary_token_ts = len(
            LargeVocabularyNotelikeTokeniser.SUPPORTED_VALUES) * note_section_size + 4 + 24

        for token in tokens:
            if token <= 3:
                pass
            elif 4 <= token <= 27:
                cur_time += token - 3
            elif 28 <= token <= boundary_token_ts - 1:
                note_cof_distance = ((token - 28) % note_section_size) % 12 - 5
                note_octave_shift = ((token - 28) % note_section_size) // 12 - 8
                note_value = LargeVocabularyNotelikeTokeniser.SUPPORTED_VALUES[(token - 28) // note_section_size]

                octave = prv_note // 12 - 1
                octave += note_octave_shift
                note_base = CircleOfFifths.from_distance(prv_note, note_cof_distance)
                note = note_base + octave * 12 + 12  # Shifts notes to A0 (21, need to increase from CoF calculation)

                if not (21 <= note <= 108):
                    LOGGER.warning(f"Invalid note: {note}")
                while note < 21:
                    note += 12
                while note > 108:
                    note -= 12

                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_ON, note=note, time=cur_time))
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_OFF, note=note, time=cur_time + note_value))
                prv_note = note
            elif boundary_token_ts <= token <= boundary_token_ts + 14:
                seq.add_absolute_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, time=cur_time,
                            numerator=token - boundary_token_ts + 2,
                            denominator=8)
                )
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq



import enum


class Note(enum.Enum):
    C = 0
    C_S = 1
    D = 2
    D_S = 3
    E = 4
    F = 5
    F_S = 6
    G = 7
    G_S = 8
    A = 9
    A_S = 10
    B = 11


class Key(enum.Enum):
    C = "C"
    G = "G"
    D = "D"
    A = "A"
    E = "E"
    B = "B"
    F_S = "F#"
    C_S = "C#"
    F = "F"
    B_B = "Bb"
    E_B = "Eb"
    A_B = "Ab"
    D_B = "Db"
    G_B = "Gb"
    C_B = "Cb"

    @staticmethod
    def transpose_key(key, transpose_by):
        if transpose_by % 12 != 0:
            if key in MusicMapping.key_transpose_mapping:
                key = MusicMapping.key_transpose_mapping[key]

            index = MusicMapping.key_transpose_order.index(key)
            index = (index + transpose_by) % 12
            return MusicMapping.key_transpose_order[index]


class CircleOfFifths:
    """Represents knowledge about the circle of fifths, note that some pitches are doubly assigned which can lead to
    ambiguity."""

    circle_of_fifths_order = [Note.C_S, Note.G_S, Note.D_S, Note.A_S, Note.F,
                              Note.C, Note.G, Note.D, Note.A, Note.E, Note.B,
                              Note.F_S]

    @staticmethod
    def get_distance(from_note_val: int, to_note_val: int):
        from_pos = CircleOfFifths.circle_of_fifths_order.index(Note(from_note_val % 12)) - 5
        to_pos = CircleOfFifths.circle_of_fifths_order.index(Note(to_note_val % 12)) - 5

        to_pos = (to_pos - from_pos) % 12

        distance_right = to_pos
        distance_left = 12 - distance_right

        if distance_left == distance_right:
            distance = distance_right
        elif distance_right < distance_left:
            distance = distance_right
        else:
            distance = -distance_left

        assert -5 <= distance <= 6

        return distance

    @staticmethod
    def from_distance(base_note_val: int, cof_distance: int):
        base_pos = CircleOfFifths.circle_of_fifths_order.index(Note(base_note_val % 12))
        return CircleOfFifths.circle_of_fifths_order[(base_pos + cof_distance) % 12].value


class MusicMapping:
    KeyKeyMapping = {"C": Key.C, "G": Key.G, "D": Key.D, "A": Key.A, "E": Key.E, "B": Key.B, "F#": Key.F_S,
                     "C#": Key.C_S,
                     "F": Key.F, "Bb": Key.B_B, "Eb": Key.E_B, "Ab": Key.A_B, "Db": Key.D_B, "Gb": Key.G_B,
                     "Cb": Key.C_B,
                     "Am": Key.C, "Em": Key.G, "Bm": Key.D, "F#m": Key.A, "C#m": Key.E, "G#m": Key.B, "D#m": Key.F_S,
                     "Dm": Key.F, "Gm": Key.B_B, "Cm": Key.E_B, "Fm": Key.A_B, "Bbm": Key.D_B, "Ebm": Key.G_B}

    KeyNoteMapping = {
        # Notes belonging to scale, accidentals
        Key.C: ([Note.C, Note.D, Note.E, Note.F, Note.G, Note.A, Note.B], 0),
        Key.G: ([Note.G, Note.A, Note.B, Note.C, Note.D, Note.E, Note.F_S], 1),
        Key.D: ([Note.D, Note.E, Note.F_S, Note.G, Note.A, Note.B, Note.C_S], 2),
        Key.A: ([Note.A, Note.B, Note.C_S, Note.D, Note.E, Note.F_S, Note.G_S], 3),
        Key.E: ([Note.E, Note.F_S, Note.G_S, Note.A, Note.B, Note.C_S, Note.D_S], 4),
        Key.B: ([Note.B, Note.C_S, Note.D_S, Note.E, Note.F_S, Note.G_S, Note.A_S], 5),
        Key.F_S: ([Note.F_S, Note.G_S, Note.A_S, Note.B, Note.C_S, Note.D_S, Note.F], 6),
        Key.C_S: ([Note.C_S, Note.D_S, Note.F, Note.F_S, Note.G_S, Note.A_S, Note.C], 7),
        Key.F: ([Note.F, Note.G, Note.A, Note.A_S, Note.C, Note.D, Note.E], 1),
        Key.B_B: ([Note.A_S, Note.C, Note.D, Note.D_S, Note.F, Note.G, Note.A], 2),
        Key.E_B: ([Note.D_S, Note.F, Note.G, Note.G_S, Note.A_S, Note.C, Note.D], 3),
        Key.A_B: ([Note.G_S, Note.A_S, Note.C, Note.C_S, Note.D_S, Note.F, Note.G], 4),
        Key.D_B: ([Note.C_S, Note.D_S, Note.F, Note.F_S, Note.G_S, Note.A_S, Note.C], 5),
        Key.G_B: ([Note.F_S, Note.G_S, Note.A_S, Note.B, Note.C_S, Note.D_S, Note.F], 6),
        Key.C_B: ([Note.B, Note.C_S, Note.D_S, Note.E, Note.F_S, Note.G_S, Note.A_S], 7)}

    key_transpose_order = [Key.C, Key.C_S, Key.D, Key.E_B, Key.E, Key.F, Key.F_S, Key.G, Key.A_B, Key.A, Key.B_B, Key.B]
    key_transpose_mapping = {Key.D_B: Key.C_S, Key.G_B: Key.F_S, Key.C_B: Key.B}



from __future__ import annotations

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.misc.music_theory import MusicMapping


class MidiMessage:

    def __init__(self, message_type=None, control=None, denominator=None, numerator=None, key=None, note=None,
                 time=None, velocity=None, program=None) -> None:
        super().__init__()
        self.message_type = message_type
        self.control = control
        self.denominator = denominator
        self.numerator = numerator
        self.key = key
        self.note = note
        self.time = time
        self.velocity = velocity
        self.program = program

    @staticmethod
    def parse_mido_message(mido_message) -> MidiMessage:
        msg = MidiMessage()

        msg.time = mido_message.time

        if mido_message.type == "note_on" and mido_message.velocity > 0:
            msg.message_type = MessageType.NOTE_ON
            msg.note = mido_message.note
            msg.velocity = mido_message.velocity
        elif (mido_message.type == "note_on" and mido_message.velocity == 0) or mido_message.type == "note_off":
            msg.message_type = MessageType.NOTE_OFF
            msg.note = mido_message.note
            msg.velocity = mido_message.velocity
        elif mido_message.type == "time_signature":
            msg.message_type = MessageType.TIME_SIGNATURE
            msg.denominator = mido_message.denominator
            msg.numerator = mido_message.numerator
        elif mido_message.type == "key_signature":
            msg.message_type = MessageType.KEY_SIGNATURE
            msg.key = MusicMapping.KeyKeyMapping[mido_message.key]
        elif mido_message.type == "control_change":
            msg.message_type = MessageType.CONTROL_CHANGE
            msg.control = mido_message.control
            msg.velocity = mido_message.value
        elif mido_message.type == "program_change":
            msg.message_type = MessageType.PROGRAM_CHANGE
            msg.program = mido_message.program

        return msg

    @staticmethod
    def parse_internal_message(message: Message) -> MidiMessage:
        return MidiMessage(message_type=message.message_type, time=message.time, note=message.note,
                           velocity=message.velocity, control=message.control, numerator=message.numerator,
                           denominator=message.denominator, key=message.key, program=message.program)

    def __str__(self) -> str:
        return f"MidiMessage(type={self.message_type}, time={self.time}, note={self.note})"



from __future__ import annotations

import copy
from statistics import geometric_mean
from typing import TYPE_CHECKING

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.exceptions.sequence_exception import SequenceException
from scoda.misc.scoda_logging import get_logger
from scoda.misc.util import binary_insort, find_minimal_distance, regress, minmax, simple_regression, \
    get_note_durations, \
    get_tuplet_durations, get_dotted_note_durations
from scoda.sequences.abstract_sequence import AbstractSequence
from scoda.settings.settings import PPQN, DIFF_DUAL_NOTE_VALUES_UPPER_BOUND, \
    DIFF_DUAL_NOTE_VALUES_LOWER_BOUND, NOTE_VALUE_UPPER_BOUND, NOTE_VALUE_LOWER_BOUND, VALID_TUPLETS, DOTTED_ITERATIONS, \
    SCALE_LOGLIKE

if TYPE_CHECKING:
    from scoda.sequences.relative_sequence import RelativeSequence


class AbsoluteSequence(AbstractSequence):
    """Class representing a sequence with absolute message timings.
    """

    LOGGER = get_logger(__name__)

    # General Methods

    def __init__(self) -> None:
        super().__init__()

    def __copy__(self) -> AbsoluteSequence:
        cpy = AbsoluteSequence()

        for message in self.messages:
            cpy.messages.append(copy.copy(message))

        return cpy

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, AbsoluteSequence):
            return False

        self_array = self.get_message_time_pairings()
        other_array = o.get_message_time_pairings()

        if len(self_array) != len(other_array):
            return False

        for self_pair, other_pair in zip(self_array, other_array):
            if self_pair[0].time != other_pair[0].time or \
                    self_pair[0].note != other_pair[0].note or \
                    self_pair[1].time != other_pair[1].time:
                return False

        return True

    # Basic Methods

    def add_message(self, msg: Message) -> None:
        """Adds the given message to the current sequence."""
        binary_insort(self.messages, msg)

    def _add_message_unsorted(self, msg: Message) -> None:
        """Adds the given message to the current sequence."""
        self.messages.append(msg)

    def cutoff(self, maximum_length, reduced_length) -> None:
        """Reduces the length of all notes longer than the maximum length to this value.

        Args:
            maximum_length: Maximum note length allowed in this sequence.
            reduced_length: The length violating notes are assigned.

        """
        note_array = self.get_message_time_pairings()

        for entry in note_array:
            if len(entry) == 1:
                if not entry[0].message_type == MessageType.NOTE_ON:
                    raise SequenceException("Cutoff: Note was closed without having been opened.")
                self.add_message(
                    Message(message_type=MessageType.NOTE_OFF, note=entry[0].note, time=entry[0].time + maximum_length))
            else:
                if entry[1].time - entry[0].time > maximum_length:
                    entry[1].time = entry[0].time + reduced_length

        self.normalise_absolute()

    def merge(self, sequences: list[AbsoluteSequence]) -> None:
        """Merges this sequence with all the given ones.

        In case of the creation of overlapping notes, these will be combined. The earliest start and the latest end will
        be used for the newly created note.

        Args:
            sequences: The sequence to merge with this one.

        """
        for sequence in sequences:
            for msg in copy.copy(sequence.messages):
                self._add_message_unsorted(msg)

        self.normalise_absolute()

    def normalise_absolute(self) -> None:
        self.sort()

    def quantise(self, step_sizes: list[int] = None) -> None:
        """Quantises the sequence to a given grid.

        Quantises the sequence stored in this object according to the given step sizes. These step sizes determine the
        size of the underlying grid, e.g. a step size of 3 would allow for messages to be placed at multiples of 3
        ticks. Note that the induced length of the notes is dependent on the `PPQN`, e.g., with a `PPQN` of 24,
        a step size of 3 would be equivalent to a grid conforming to thirty-second notes. If there exists a tie between
        two grid boundaries, these are first resolved by whether the quantisation would prevent a note-length of 0,
        then by the order of the `step_sizes` array. The result of this operation is that all messages of this
        sequence have a time divisible by one of the values in `step_sizes`. If the quantisation resulted in two
        notes overlapping, the second note will be removed. See `scoda.utils.utils.get_note_durations`,
        `scoda.utils.utils.get_tuplet_durations` and `scoda.utils.utils.get_dotted_note_durations` for generating the
        `step_sizes` array.

        Args:
            step_sizes: Array of numbers corresponding to divisors of the grid length.

        """
        if step_sizes is None:
            quantise_parameters = get_note_durations(1, 8)
            quantise_parameters += get_tuplet_durations(quantise_parameters, 3, 2)
            step_sizes = quantise_parameters

        # List of finally quantised messages
        quantised_messages = []
        # Keep track of open messages, in order to guarantee quantisation does not smother them
        open_messages = dict()
        # Keep track of from when to when notes are played, in order to eliminate double notes
        message_timings = dict()

        for msg in self.messages:
            message_original_time = msg.time
            message_to_append = copy.copy(msg)

            # Positions the note would land at according to each of the quantisation parameters
            positions_left = [(message_original_time // step_size) * step_size for step_size in step_sizes]
            positions_right = [positions_left[i] + step_sizes[i] for i in range(0, len(step_sizes))]

            possible_positions = positions_left + positions_right
            valid_positions = []

            # Consider quantisations that could smother notes
            if msg.message_type == MessageType.NOTE_ON:
                valid_positions += possible_positions
                message_to_append.time = valid_positions[find_minimal_distance(message_original_time, valid_positions)]

                # Check if note was not yet closed
                if msg.note in open_messages:
                    AbsoluteSequence.LOGGER.warning(f"Quantisation: Note {msg.note} not previously stopped.")
                    quantised_messages.append(
                        Message(message_type=MessageType.NOTE_OFF, note=msg.note, time=message_to_append.time))
                    open_messages.pop(msg.note, None)
                    message_timings[msg.note].append(message_to_append.time)

                # Check if we can open note without overlaps
                if msg.note not in message_timings \
                        or not message_to_append.time < message_timings[msg.note][1]:
                    open_messages[msg.note] = message_to_append.time
                    message_timings[msg.note] = [message_to_append.time]
                # In this case note would overlap with other, existing note
                else:
                    message_to_append = None
            elif msg.message_type == MessageType.NOTE_OFF:
                # Message is currently open, have to quantize
                if msg.note in open_messages:
                    note_open_timing = open_messages.pop(msg.note, None)

                    # Add possible positions for stop messages, making sure the belonging note is not smothered
                    for position in possible_positions:
                        if not position - note_open_timing <= 0:
                            valid_positions.append(position)

                    # If no valid positions exists, set note length to 0
                    if len(valid_positions) == 0:
                        valid_positions.append(note_open_timing)

                    # Valid positions will always exist, since if order held before quantisation, same will hold
                    # after, and if initially no valid position was found note length will be set to 0
                    message_to_append.time = valid_positions[
                        find_minimal_distance(message_original_time, valid_positions)]
                    message_timings[msg.note].append(message_to_append.time)

                # Message is not currently open (e.g., if start message was removed due to an overlap)
                else:
                    message_to_append = None
            else:
                valid_positions += possible_positions
                message_to_append.time = valid_positions[find_minimal_distance(message_original_time, valid_positions)]

            if message_to_append is not None:
                quantised_messages.append(message_to_append)

        # Remove smothered notes
        message_timings_with_indices = dict()
        original_indices_to_remove = []

        # Get indices of violating messages
        for i, msg in enumerate(quantised_messages):
            if msg.message_type == MessageType.NOTE_ON:
                message_timings_with_indices[msg.note] = (i, msg.time)
            elif msg.message_type == MessageType.NOTE_OFF:
                j, time = message_timings_with_indices.pop(msg.note)
                if msg.time - time <= 0:
                    original_indices_to_remove.extend([j, i])

        # Remove messages
        for index_shifter, index_to_remove in enumerate(original_indices_to_remove):
            quantised_messages.pop(index_to_remove - index_shifter)

        self.messages = quantised_messages
        self.normalise_absolute()

    def quantise_note_lengths(self, possible_durations=None, standard_length=PPQN, do_not_extend=False) -> None:
        """Quantises the note lengths of this sequence, only affecting the ending of the notes.

        Quantises notes to the given values, ensuring that all notes are of one of the sizes defined by the
        parameters. See `scoda.utils.utils.get_note_durations`, `scoda.utils.utils.get_tuplet_durations` and
        `scoda.utils.utils.get_dotted_note_durations` for generating the `possible_durations` array. Tries to shorten
        or extend the end of each note in such a way that the note duration is exactly one of the values given in
        `possible_durations`. If this is not possible (e.g., due to an overlap with another note that would occur),
        the note that can neither be shortened nor lengthened will be removed from the sequence. Note that this is
        only the case if the note was shorter than the smallest legal duration specified, and thus cannot be shortened.

        Args:
            possible_durations: An array containing exactly the valid note durations in ticks.
            standard_length: Note length for notes which are not closed.
            do_not_extend: Determines if notes are only allowed to be shortened, e.g., for bars.

        """
        # Construct current durations
        notes = self.get_message_time_pairings(standard_length=standard_length)
        # Track when each type of note occurs, in order to check for possible overlaps
        note_occurrences = dict()
        quantised_messages = []

        # Construct possible durations
        if possible_durations is None:
            normal_durations = get_note_durations(NOTE_VALUE_UPPER_BOUND, NOTE_VALUE_LOWER_BOUND)
            triplet_durations = []
            for valid_tuplet in VALID_TUPLETS:
                triplet_durations.extend(get_tuplet_durations(normal_durations, valid_tuplet[0], valid_tuplet[1]))
            dotted_durations = get_dotted_note_durations(normal_durations, DOTTED_ITERATIONS)
            possible_durations = normal_durations + triplet_durations + dotted_durations

        # Construct array keeping track of when each note occurs
        for pairing in notes:
            note = pairing[0].note
            note_occurrences.setdefault(note, [])
            note_occurrences[note].append(pairing)

        # Handle each note, pairing consists of start and stop message
        for i, pairing in enumerate(notes):
            note = pairing[0].note
            current_duration = pairing[1].time - pairing[0].time
            valid_durations = copy.copy(possible_durations)

            # Check if the current note is not the last note, in this case clashes with a next note could exist
            index = note_occurrences[note].index(pairing)
            if index != len(note_occurrences[note]) - 1:
                possible_next_pairing = note_occurrences[note][index + 1]

                # Possible durations contains the same as valid durations at the beginning
                for possible_duration in possible_durations:
                    possible_correction = possible_duration - current_duration

                    # If we cannot extend the note, remove the time from possible times
                    if pairing[1].time + possible_correction > possible_next_pairing[0].time:
                        valid_durations.remove(possible_duration)

            # If we are not allowed to extend note lengths, remove all positive corrections
            for possible_duration in possible_durations:
                possible_correction = possible_duration - current_duration

                if possible_correction > 0 and do_not_extend and possible_duration in valid_durations:
                    valid_durations.remove(possible_duration)

            # Check if we have to remove the note
            if len(valid_durations) == 0:
                notes[i] = []
            else:
                current_duration = pairing[1].time - pairing[0].time
                best_fit = valid_durations[find_minimal_distance(current_duration, valid_durations)]
                correction = best_fit - current_duration
                pairing[1].time += correction

        for pairing in notes:
            quantised_messages.extend(pairing)

        for msg in self.messages:
            if msg.message_type is not MessageType.NOTE_ON and msg.message_type is not MessageType.NOTE_OFF:
                quantised_messages.append(msg)

        self.messages = quantised_messages
        self.normalise_absolute()

    def sort(self) -> None:
        """Sorts the sequence according to the timings of the messages.

        This sorting procedure is stable, if two messages occurred in a specific order at the same time before the sort,
        they will occur in this order after the sort.

        """
        self.messages.sort(key=lambda x: (x.time, x.message_type, x.note))

    # Misc. Methods

    def get_message_time_pairings(self, message_types: list[MessageType] = None, standard_length=PPQN,
                                  impute_notes=True) -> list[
        list[Message]]:
        """Creates an ordered list containing lists of messages of the given types.
        These lists contain all entries for messages that belong together, e.g., open and close note messages.

        Args:
            message_types: All message types to include.
            standard_length: The length used for notes that have not been closed.
            impute_notes: If unclosed notes should be closed and unopened ones should be ignored.

        Returns: An array of tuples of two messages constituting a note.

        """
        if message_types is None:
            message_types = [MessageType.NOTE_ON, MessageType.NOTE_OFF]

        self.normalise_absolute()

        open_messages = dict()
        notes: [[]] = []
        i = 0

        # Collect notes
        for msg in self.messages:
            if msg.message_type in message_types:
                # Add notes to open messages
                if msg.message_type == MessageType.NOTE_ON:
                    if msg.note in open_messages and impute_notes:
                        AbsoluteSequence.LOGGER.warning(
                            f"Time Pairings: Note {msg.note} at time {msg.time} not previously stopped.")
                        index = open_messages.pop(msg.note)
                        notes[index].append(Message(message_type=MessageType.NOTE_OFF, note=msg.note, time=msg.time))

                    open_messages[msg.note] = i
                    notes.insert(i, [msg])
                    i += 1

                # Add closing message to fitting open message
                elif msg.message_type == MessageType.NOTE_OFF:
                    if msg.note not in open_messages and impute_notes:
                        AbsoluteSequence.LOGGER.warning(
                            f"Time Pairings: Note {msg.note} at time {msg.time} not previously started.")
                    else:
                        index = open_messages.pop(msg.note)
                        notes[index].append(msg)

                else:
                    notes.insert(i, [msg])
                    i += 1

        # Check unclosed notes
        for pairing in notes:
            if len(pairing) == 1 and pairing[0].message_type == MessageType.NOTE_ON and impute_notes:
                pairing.append(Message(message_type=MessageType.NOTE_OFF, time=pairing[0].time + standard_length))

        return notes

    def get_message_timings_of_type(self, message_types: list[MessageType]) -> list[tuple[int, Message]]:
        """Searches for messages that fit one of the given types.

        Args:
            message_types: Which message types to search for.

        Returns: A list containing the found messages and their absolute points in time.

        """
        timings = []

        for msg in self.messages:
            if msg.message_type in message_types:
                timings.append((msg.time, msg))

        return timings

    def get_sequence_duration(self) -> int:
        """Calculates the overall duration of this sequence, given in ticks.

        Returns: The duration of this sequence.

        """
        return self.messages[-1].time

    def to_relative_sequence(self) -> RelativeSequence:
        """Converts this AbsoluteSequence to a RelativeSequence.

        Returns: The relative representation of this sequence.

        """
        from scoda.sequences.relative_sequence import RelativeSequence
        relative_sequence = RelativeSequence()
        current_point_in_time = 0

        for msg in self.messages:
            time = msg.time
            # Check if we have to add wait messages
            if time > current_point_in_time:
                relative_sequence.add_message(
                    Message(message_type=MessageType.WAIT, time=time - current_point_in_time))
                current_point_in_time = time

            if msg.message_type != MessageType.INTERNAL:
                message_to_add = copy.copy(msg)
                message_to_add.time = None
                relative_sequence.add_message(message_to_add)

        return relative_sequence

    # Difficulty Methods

    def diff_note_values(self) -> float:
        """Calculates complexity of the piece regarding the geometric mean of the note values.

        Calculates the geometric mean based on all occurring notes in this sequence and then applies linear scaling
        to it. Returns a value from 0 to 1, where 0 indicates low difficulty. If no notes exist in this sequence,
        the lowest complexity rating is returned.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        notes = self.get_message_time_pairings()
        durations = []

        for pairing in notes:
            durations.append(pairing[1].time - pairing[0].time)

        if len(durations) == 0:
            durations.append(DIFF_DUAL_NOTE_VALUES_LOWER_BOUND)

        mean = geometric_mean(durations)
        bound_mean = minmax(0, 1,
                            simple_regression(DIFF_DUAL_NOTE_VALUES_UPPER_BOUND, 1, DIFF_DUAL_NOTE_VALUES_LOWER_BOUND,
                                              0, mean))

        return minmax(0, 1, bound_mean)

    def diff_rhythm(self) -> float:
        """Calculates difficulty based on the rhythm of the sequence.

        For this calculation, note values are weighted by checking if they are normal values, dotted, or tuplet ones.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        notes = self.get_message_time_pairings()

        # If sequence is empty, return easiest difficulty
        if len(notes) == 0:
            return 0

        note_durations = get_note_durations(NOTE_VALUE_UPPER_BOUND, NOTE_VALUE_LOWER_BOUND)

        tuplets = []
        for tuplet_duration in VALID_TUPLETS:
            tuplets.append(get_tuplet_durations(note_durations, tuplet_duration[0], tuplet_duration[1]))

        dotted_durations = get_dotted_note_durations(note_durations, DOTTED_ITERATIONS)

        notes_regular = []
        notes_dotted = []
        notes_tuplets = []

        for note in notes:
            duration = note[1].time - note[0].time

            if duration in note_durations:
                notes_regular.append(note)
            elif any(duration in x for x in tuplets):
                notes_tuplets.append(note)
            elif duration in dotted_durations:
                notes_dotted.append(note)
            else:
                AbsoluteSequence.LOGGER.warning(f"Difficulty Rhythm: Note value {duration} not in known values.")

        rhythm_occurrences = 0

        rhythm_occurrences += len(notes_dotted) * 0.5
        rhythm_occurrences += len(notes_tuplets) * 1

        unscaled_difficulty = minmax(0, 1, rhythm_occurrences / len(notes))
        scaled_difficulty = regress(unscaled_difficulty, SCALE_LOGLIKE)

        return minmax(0, 1, scaled_difficulty)



from __future__ import annotations

import enum


class TokenisationFlags(enum.Enum):
    RUNNING_VALUE = "running_value"
    RUNNING_PITCH = "running_pitch"
    RUNNING_OCTAVE = "running_octave"
    RUNNING_TIME_SIG = "running_time_signature"



import logging
import logging.config

loggers: dict = dict()


def setup():
    global loggers

    logger = logging.getLogger("scoda")
    loggers["scoda"] = logger

    logger.setLevel(logging.INFO)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)

    log_format = logging.Formatter("%(asctime)s - [%(name)s] - %(threadName)s - %(levelname)s: %(message)s")
    console_handler.setFormatter(log_format)

    logger.addHandler(console_handler)
    logger.propagate = False


def get_logger(logger_designation: str = "scoda") -> logging.Logger:
    if logger_designation is None:
        logger_designation = "scoda"

    if logger_designation not in loggers:
        loggers[logger_designation] = logging.getLogger(logger_designation)

    return loggers[logger_designation]


setup()
